
from Npp import *
import socket
import threading
import xml.etree.ElementTree as ET

# module/global scope is a read-only shadow within the local function/method scope: you can access it,
# providing there's nothing declared with the same name in local scope, but you can't change it.
# So we can access variables declared in the top-level in functions, but need the global keyword if we want to change them
# in a function.

# Show the PythonScript console
console.show()
console.clear()

DEBUG       = True

sock        = None
t_stop      = threading.Event()

bp_marker   = None
breakpoints = set()
indicator = 10
margin = 3

connected   = False
suspended   = True

#this_filename = notepad.getCurrentFilename()
#out = notepad.new("out")
#notepad.activateFile(this_filename)

# A couple output methods for hopeless Ruby folks.
def puts(s):
    console.write("%s" % s)

def dputs(s):
    if DEBUG:
        puts(s)


############################################################
# Breakpoints
############################################################

# bp = Breakpoint({'location': 'C:/Users/Jim/AppData/Roaming/SketchUp/SketchUp 2014/SketchUp/Plugins/jf-foo.rb:7', 'no': '1'})
# bp.no       - returns breakpoint number (int) assigned by debugger
# bp.line     - returns line number (int) of breakpoint
# bp.file - returns the filename (str) of the breakpoint
class Breakpoint:
    def __init__(self, data):
        self.data = data
        self.no       = int( data['no'] )
        self.location = data['location']
        self.split_location()

    # split the location into file and line number
    def split_location(self):
        tmp_string       = self.location
        last_colon_index = tmp_string.rindex(':')
        self.file        = tmp_string[:last_colon_index]
        self.line        = int( tmp_string[last_colon_index+1:] )

    def __repr__(self):
        return "Breakpoint: no:%s line:%s file:%s" % (self.no, self.line, self.file)

def find_breakpoint_by_no(n):
    found = None
    for bp in breakpoints:
        if bp.no == n:
            found = bp
            break
    return found

def find_breakpoint_by_line(line):
    found = None
    for bp in breakpoints:
        if bp.line == line:
            found = bp
            break
    return found

def list_breakpoints():
    for bp in breakpoints:
        txt = "  %s %s:%s\n" % (bp.no, bp.file, bp.line)
        console.write(txt)

def clear_breakpoints():
    bpts = breakpoints.copy()
    for bp in bpts:
        console.write("%s\n" % repr(bp))
        req_bp_del(bp)


############################################################
# Debugger
############################################################

def init():
    connected = connect()
    if connected:
        start_listener()
        #start()
    else:
        console.write("Could not connect to SketchUp.")

def connect():
    console.write("Connecting to SketchUp..\n")
    global sock
    sock    = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    address = ('localhost', 1234)
    try:
        sock.connect(address)
    except socket.error, exc:
        if sock:
            sock.close()
        console.write("socket error: %s\n" % exc)
        connected = False
    else:    
    # are we really connected? SketchUp debugger does not send a reply either way.
        connected = True
    return connected

def close():
    global sock
    if DEBUG:
        console.write("closing socket.\n")
    sock.close()
        
# we run the rec() function in a thread so it does not block
# t_stop event is used to stop the thread.
def start_listener():
    console.write("Starting Listener..\n")
    global t_stop
    t_stop.clear()
    t = threading.Thread(target=rec, args=((t_stop,)))
    t.start()


# rec() runs in own thread.
# Needs to handle more data than 4096 since 'var global' command returns a lot of data.
def rec(stop_event):
    #global sock
    if DEBUG:
        console.write("stop_event: %s\n" % repr(stop_event))
        
    while( not stop_event.is_set() ):
        try:
            data = sock.recv(65535)
            data = data.strip()
            handle_msg(data)
        except socket.error, (value, message):
            console.write("socket error: %s" % message)
            sock.close()
            return


# Send a message to the debug server
def send(msg):
    #global sock
    if msg[-1] != '\n':
        msg += '\n'
    if DEBUG:
        console.write("sending: " + msg)
    sock.send(msg)


# handle message response from debug server
# data from debugger is xml
def handle_msg(msg):

    # view the raw xml from SU DB
    if DEBUG:
        for line in msg.split("\n"):
            console.write("%s\n" % line)

    try:
        tree = ET.fromstring(msg)
    except ET.ParseError:
        console.write("ParseError:\n")
        console.write(repr(msg))
        console.write("\n")
        return

    tag_name = tree.tag
    #console.write(repr(tree.attrib))
    #console.write(repr(tree.keys()))
    #console.write(repr(tree.items()))

    if tag_name == 'breakpointAdded':
        breakpoint_added(tree.attrib)

    elif tag_name == 'breakpointDeleted':
        breakpoint_deleted(tree.attrib)

    # occurs when a breakpoint is hit in SketchUp
    elif tag_name == 'breakpoint':
        handle_breakpoint(tree.attrib)

    # v l or v g command
    elif tag_name == 'variables':
        list_variables(tree)

    # occurs after a step command
    elif tag_name == 'suspended':
        suspend(tree.attrib)

    elif tag_name == 'frames':
        show_frames(tree)

    else:
        console.write("unhandled tag_name:%s\n" % repr(tag_name))

def breakpoint_added(d):
    bp = Breakpoint(d)
    add_breakpoint(bp)

def add_breakpoint(bp):
    #editor.setReadOnly(True)
    add_marker(bp.line)
    breakpoints.add(bp)

def breakpoint_deleted(d):
    no = int(d['no'])
    delete_breakpoint(no)

def delete_breakpoint(no):
    bp = find_breakpoint_by_no(no)
    delete_marker(bp.line)
    breakpoints.remove(bp)


def handle_breakpoint(d):
    line = int( d['line'] )
    console.write("stopped at breakpoint on line %s\n" % line)
    global suspended
    suspended = True
    clear_ind()
    ind_line(line - 1)
    where()
    local_vars()
    #editor.gotoLine(line - 1)

# Occurs on step command
def suspend(d):
    global suspended
    suspended = True
    console.write("suspended: %s\n" % d)
    line = int( d['line'] )
    console.write("Suspended at line %s\n" % line)
    #editor.gotoLine(line - 1)
    where()
    clear_ind()
    ind_line(line - 1)


def list_variables(tree):
    console.write("Variables:\n")
    for child in tree:
        console.write('\t')
        console.write(child.attrib['name'])
        console.write(' => ')
        console.write(repr(child.attrib['value']))
        console.write('\n')

def show_frames(tree):
    console.write("Frames:\n")
    for child in tree:
        if "current" in child.attrib:
            current = child.attrib["current"] == "yes"
        else:
            current = False
        no = child.attrib["no"]
        line = child.attrib["line"]
        file = child.attrib["file"]
        if current:
            console.write("  * %s %s:%s\n" % (no, file, line))
            show_file(file, line)
        else:
            console.write("    %s %s:%s\n" % (no, file, line))


def show_file(file, line):
    notepad.open(file)
    notepad.activateFile(file)
    editor.setFirstVisibleLine(int(line)-5)
#
# Debugger Protocol Commands
#

def b(filename, linenum):
    cmd = "break %s:%s\n" % (filename, linenum)
    send(cmd)

def cont():
    # run until next breakpoint or program ends
    send("cont")
    clear_ind()
    suspended = False

def start():
    send("start")
    clear_ind()
    suspended = False

def exit():
    # https://github.com/SketchUp/sketchup-ruby-debugger/blob/master/DebugServer/UI/RDIP/RDIP.cpp#L238
    # // Stop debugging. First let SU continue in case it's at a breakpoint.
    # std::lock_guard<std::mutex> lock(server_wait_mutex_);
    # server_can_continue_ = true;
    # server_wait_cond_.notify_all();
    # // Now call Stop. It's unclear if it is ok to do this from the RDIP thread
    # // but it appears to work.
    # server_->Stop();
    global connected
    send("exit")
    connected = False

# Run until we return from the method
def finish():
    # https://github.com/SketchUp/sketchup-ruby-debugger/blob/master/DebugServer/UI/RDIP/RDIP.cpp#L285
    # server_->StepOut();
    # std::lock_guard<std::mutex> lock(server_wait_mutex_);
    # server_can_continue_ = true;
    # server_wait_cond_.notify_all();
    clear_ind()
    send("finish")

# Step Into
def step():
    global suspended
    suspended = False
    clear_ind()
    send("step")

def local_vars():
    send("v l")

def where():
    send("where")

def next():
    #server_->StepOver();
    #std::lock_guard<std::mutex> lock(server_wait_mutex_);
    #server_can_continue_ = true;
    #server_wait_cond_.notify_all();
    clear_ind()
    send("next")

def thread_list():
    send("thread list\n")
        
# frame [0-9]+
# v inspect
# var global, v g
# var instance, v i *


############################################################
## EDITOR 
############################################################

def toggle_breakpoint(args):
    if DEBUG:
        console.write("args: %s\n" % args)
    # bug where marginclick does not have all params
    if len(args) == 2:
        return

    #if 'modifiers' in args:
        #if args['modifiers'] != KEYMOD.CTRL:
            #return

    if 'margin' in args:
        margin = args["margin"]
    else:
        margin = None
        
    if margin != 3:
        console.write("Margin != 3\n")
        return
        
    if 'position' in args:
        position = args["position"]
    else:
        position = None

    file = str(notepad.getCurrentFilename().replace("\\", "/"))
    if DEBUG:
        console.write("file: %s\n" % file)

    line = get_line(position)
    if DEBUG:
        console.write("line: %s\n" % line)
    
    #line_txt = editor.getCurLine()
    line_txt = editor.getLine(line)
    if DEBUG:
        console.write("line_txt: %s\n" % line_txt.strip())
        
    if has_marker(line, bp_marker) == True:
        bp  = find_breakpoint_by_line(line+1)
        req_bp_del(bp)
        delete_marker(line)
    else:
        line = str(line+1)
        #console.write(repr(line_txt))
        if line_txt.strip() == "":
            console.write("Will not add breakpoint to empty line:%s\n" % line)
        elif line_txt.strip()[0] == '#':
            console.write("Will not add breakpoint to comment line:%s\n" % line)
        else:
            req_bp_add(file, line)

def req_bp_del(bp):
    msg = "del " + str(bp.no)
    send(msg)

def req_bp_add(file, line):
    msg = "break " + file + ":" + line
    send(msg)

def get_first_unused_marker():
    for i in range(0, 23):
        md = editor.markerSymbolDefined(i)
        if md == 0:
            break
    md = i
    return md


def get_line(position = None):
    if position == None:
        position = editor.getCurrentPos()

    line = editor.lineFromPosition(position)
    return line


# TODO:  Should return a list of all marker numbers for line
# or separate all_markers() function.
def has_marker(line, marker):
    markers = all_markers(line)
    return (marker in markers)

def all_markers(line):
    all_markers = []
    markers = editor.markerGet(line)
    for m in range(0, 32):
        mask = 1 << m
        if markers & mask:
            all_markers.append(m)
    return all_markers

def add_marker(line):
    global bp_marker
    editor.markerAdd(line-1, bp_marker)

def delete_marker(line):
    editor.markerDelete(line-1, bp_marker)


def ind_line(line):
    pos    = editor.positionFromLine(line)
    length = editor.lineLength(line)
    editor.scrollRange(pos-1, pos+1)
    editor.setIndicatorCurrent(indicator)
    editor.indicatorFillRange(pos, length)

def clear_ind():
    editor.setIndicatorCurrent(indicator)
    editor.indicatorClearRange(0, editor.getLength())

def margin_click_handler(arg):
    # {'position': 34, 'code': 2010, 'margin': 3, 'modifiers': 0}
    #console.write("Margin Click <%s>\n" % arg)
    console.write("margin: %s\n" % arg["margin"])
    
def setup_editor():
    global bp_marker
    global indicator
    global margin
    editor.clearCallbacks()
    editor.markerDeleteAll(-1)

    # This callback does not appear to be called.
    # https://sourceforge.net/p/npppythonscript/discussion/1188886/thread/1e68f81a/#8d84
    #notepad.callback(shutdown, [NOTIFICATION.SHUTDOWN])

    ## Marker
    bp_marker = get_first_unused_marker()
    if DEBUG:
        console.write("bp_marker: %s" % bp_marker)

    editor.markerDefine(bp_marker, MARKERSYMBOL.SHORTARROW)
    #editor.markerDefine(bp_marker, MARKERSYMBOL.CIRCLE)
    editor.markerSetFore(bp_marker, (0, 0, 0))
    editor.markerSetBack(bp_marker, (255, 0, 0))

    ## Margin
    #margin = 0
    # allow markers in margin
    #editor.setMarginMaskN(margin, 33554431)
    editor.setMarginMaskN(margin, 0b11)
    #editor.setMarginMaskN(margin, 0xffffffff)
    # set margin width in pixels
    editor.setMarginWidthN(margin, 22)
    # allow margin to respond to click notifications
    editor.setMarginSensitiveN(margin, True)
    editor.callback(toggle_breakpoint, [SCINTILLANOTIFICATION.MARGINCLICK])
    #editor.callback(margin_click_handler, [SCINTILLANOTIFICATION.MARGINCLICK])

    # Indicator - highlight lines
    indicator = 10
    editor.setIndicatorCurrent(indicator)
    editor.indicSetStyle(indicator,  INDICATORSTYLE.ROUNDBOX) # box
    editor.indicSetFore(indicator, (255, 0, 0))
    #editor.setIndicatorValue(5)


setup_editor()

